from . import aliases
import json, loady, os

ISNT_GIT_PATH_ERROR = """\
Because the --isolate flag is set, all paths must start with //git.
Your path was %s."""

guess_name = loady.importer.guess_name


def data(name):
    if not name:
        return {}

    try:
        return json.loads(name)
    except:
        return loady.data.load(name, True)


def code(name, python_path=None):
    return name and loady.code.load_code(name, python_path)


def module(name, python_path=None):
    return name and loady.code.load_module(name, python_path)


def extender(path):
    parts = path.split(':')

    if aliases.ISOLATE:
        if not all(x.startswith('//git/') for x in parts):
            raise ValueError(ISNT_GIT_PATH_ERROR % path)
    else:
        parts.insert(0, os.getcwd())

    return parts and loady.sys_path.extender(':'.join(parts))



import copy
from . import attributes, construct, cleanup, defaults, load, recurse
from .. util import exception


class Project:
    CHILDREN = 'maker', 'drivers', 'layout', 'animation'

    @staticmethod
    def pre_recursion(desc):
        return cleanup.cleanup(desc)

    def construct_child(self, datatype, typename=None, **kwds):
        construct = getattr(datatype, 'construct', None)
        if construct:
            return construct(self, **kwds)
        return datatype(**kwds)

    def __init__(self, *, drivers, layout, maker, path, animation, **kwds):
        attributes.check(kwds, 'project')
        self.path = path
        layout = layout or cleanup.cleanup_layout(animation)

        self.maker = self.construct_child(**maker)

        def post(desc):
            return self.construct_child(**desc)

        def create(root, name):
            with exception.add('Unable to create ' + name):
                return recurse.recurse(
                    root,
                    pre=None,
                    post=post,
                    python_path='bibliopixel.' + name)

        self.drivers = [create(d, 'drivers') for d in drivers]
        with exception.add('Unable to create layout'):
            self.layout = self.construct_child(**layout)
        self.animation = create(animation, 'animation')

    def make_animation(self):
        return self.animation


def project(*descs, **kwds):
    desc = defaults.merge(*descs, **kwds)
    with load.extender(desc.get('path', '')):
        desc = recurse.recurse(desc)
        return construct.construct(**desc)


def read_project(location, threaded=None, default=None, **kwds):
    project_data = load.data(location)
    if threaded is not None:
        project_data.setdefault('run', {})['threaded'] = threaded

    return project(default, copy.deepcopy(project_data), **kwds)



import glob, numbers, os, sys
from .. util import colors, log
from .. layout import Matrix

try:
    from PIL import Image, ImageSequence
except:
    Image, ImageSequence = None, None


def show_image(setter, width, height,
               image_path='', image_obj=None, offset=(0, 0), bgcolor=colors.Off,
               brightness=255):
    """Display an image on a matrix."""
    bgcolor = colors.color_scale(bgcolor, brightness)

    img = image_obj
    if image_path and not img:
        if not Image:
            error = "Please install Python Imaging Library: pip install pillow"
            log.error(error)
            raise ImportError(error)
        img = Image.open(image_path)
    elif not img:
        raise ValueError('Must provide either image_path or image_obj')

    w = min(width - offset[0], img.size[0])
    h = min(height - offset[1], img.size[1])
    ox = offset[0]
    oy = offset[1]

    for x in range(ox, w + ox):
        for y in range(oy, h + oy):
            r, g, b, a = (0, 0, 0, 255)
            rgba = img.getpixel((x - ox, y - oy))

            if isinstance(rgba, int):
                raise ValueError('Image must be in RGB or RGBA format!')
            if len(rgba) == 3:
                r, g, b = rgba
            elif len(rgba) == 4:
                r, g, b, a = rgba
            else:
                raise ValueError('Image must be in RGB or RGBA format!')

            if a == 0:
                r, g, b = bgcolor
            else:
                r, g, b = colors.color_scale((r, g, b), a)

            if brightness != 255:
                r, g, b = colors.color_scale((r, g, b), brightness)

            setter(x, y, (r, g, b))


def showImage(layout, imagePath="", imageObj=None, offset=(0, 0), bgcolor=colors.Off, brightness=255):
    """Display an image on the matrix"""
    if not isinstance(layout, Matrix):
        raise RuntimeError("Must use Matrix with showImage!")

    layout.all_off()

    return show_image(layout.set, layout.width, layout.height, imagePath, imageObj,
                      offset, bgcolor, brightness)


def loadImage(layout, imagePath="", imageObj=None, offset=(0, 0), bgcolor=colors.Off, brightness=255):
    """Display an image on the matrix"""

    if not isinstance(layout, Matrix):
        raise RuntimeError("Must use Matrix with loadImage!")

    texture = [[colors.Off for x in range(layout.width)] for y in range(layout.height)]

    def setter(x, y, pixel):
        if y >= 0 and x >= 0:
            texture[y][x] = pixel

    show_image(setter, layout.width, layout.height, imagePath, imageObj,
               offset, bgcolor, brightness)

    return texture


def convert_mode(image, mode='RGB'):
    """Return an image in the given mode."""
    return image if (image.mode == mode) else image.convert(mode=mode)


def image_to_colorlist(image, container=list):
    """Given a PIL.Image, returns a ColorList of its pixels."""
    return container(convert_mode(image).getdata())


def animated_gif_to_colorlists(image, container=list):
    """Given an animated GIF, return a list with a colorlist for each frame."""
    it = ImageSequence.Iterator(image)
    return [image_to_colorlist(i, container) for i in it]


def crop(image, top_offset=0, left_offset=0, bottom_offset=0, right_offset=0):
    """Return an image cropped on top, bottom, left or right."""
    if bottom_offset or top_offset or left_offset or right_offset:
        width, height = image.size
        box = (left_offset, top_offset,
               width - right_offset, height - bottom_offset)
        image = image.crop(box=box)

    return image


def resize(image, x, y, stretch=False, top=None, left=None, mode='RGB',
           resample=None):
    """Return an image resized."""
    if x <= 0:
        raise ValueError('x must be greater than zero')
    if y <= 0:
        raise ValueError('y must be greater than zero')

    resample = Image.ANTIALIAS if resample is None else resample
    if not isinstance(resample, numbers.Number):
        try:
            resample = getattr(Image, resample.upper())
        except:
            raise ValueError("(1) Didn't understand resample=%s" % resample)
        if not isinstance(resample, numbers.Number):
            raise ValueError("(2) Didn't understand resample=%s" % resample)

    size = x, y
    if stretch:
        return image.resize(size, resample=resample)
    result = Image.new(mode, size)

    ratios = [d1 / d2 for d1, d2 in zip(size, image.size)]
    if ratios[0] < ratios[1]:
        new_size = (size[0], int(image.size[1] * ratios[0]))
    else:
        new_size = (int(image.size[0] * ratios[1]), size[1])

    image = image.resize(new_size, resample=resample)
    if left is None:
        box_x = int((x - new_size[0]) / 2)
    elif left:
        box_x = 0
    else:
        box_x = x - new_size[0]

    if top is None:
        box_y = int((y - new_size[1]) / 2)
    elif top:
        box_y = 0
    else:
        box_y = y - new_size[1]

    result.paste(image, box=(box_x, box_y))
    return result



# Original code by msurguy: https://github.com/ManiacalLabs/BiblioPixel/issues/51#issuecomment-228662943

from . channel_order import ChannelOrder
from . driver_base import DriverBase
from .. util import log
from .. colors import gamma

WS_ERROR = """
PiWS281X Requires the rpi_ws281x C extension.

Install rpi_ws281x with the following shell commands:

    git clone https://github.com/jgarff/rpi_ws281x.git
    cd rpi_ws281x

    sudo apt-get install python-dev swig scons
    sudo scons

    cd python
    # If using default system python3
    sudo python3 setup.py build install
    # If using virtualenv, enter env then run
    python setup.py build install
"""

try:
    from neopixel import Adafruit_NeoPixel, Color as NeoColor
except:
    Color = None


PIN_CHANNEL = {
    12: 0,
    18: 0,
    40: 0,
    52: 0,
    13: 1,
    19: 1,
    41: 1,
    45: 1,
    53: 1,
    10: 0,  # Technically SPI
}

STRIP_TYPES = {
    3: 0x00100800,
    4: 0x18100800,
}


class PiWS281X(DriverBase):
    """
    Driver for controlling WS281X LEDs via the rpi_ws281x C-extension.
    Only supported on the Raspberry Pi 2 & 3
    """
    def __init__(
            self, num, gamma=gamma.NEOPIXEL, c_order=ChannelOrder.RGB, gpio=18,
            ledFreqHz=800000, ledDma=5, ledInvert=False,
            color_channels=3, brightness=255, **kwds):
        """
        num - Number of LED pixels.
        gpio - GPIO pin connected to the pixels (must support PWM! GPIO 13 or 18 (pins 33 or 12) on RPi 3).
        ledFreqHz - LED signal frequency in hertz (800khz or 400khz)
        ledDma - DMA channel to use for generating signal (Between 1 and 14)
        ledInvert - True to invert the signal (when using NPN transistor level shift)
        """
        if not Color:
            raise ValueError(WS_ERROR)
        super().__init__(num, c_order=c_order, gamma=gamma, **kwds)
        self.gamma = gamma
        if gpio not in PIN_CHANNEL.keys():
            raise ValueError('{} is not a valid gpio option!')
        try:
            strip_type = STRIP_TYPES[color_channels]
        except:
            raise ValueError('In PiWS281X, color_channels must be either 3 or 4')

        self._strip = Adafruit_NeoPixel(
            num, gpio, ledFreqHz, ledDma, ledInvert, brightness,
            PIN_CHANNEL[gpio], strip_type)
        # Intialize the library (must be called once before other functions).
        self._strip.begin()

    def set_brightness(self, brightness):
        self._strip.setBrightness(brightness)
        return True

    def _compute_packet(self):
        self._render()
        data = self._buf
        self._packet = [tuple(data[(p * 3):(p * 3) + 3]) for p in range(len(data) // 3)]

    def _send_packet(self):
        for i, p in enumerate(self._packet):
            self._strip.setPixelColor(i, NeoColor(*p))

        self._strip.show()



import argparse, os, sys
from . import common_flags
from .. util import log
from .. project.importer import import_module
from .. project import aliases


__all__ = ['main']
COMMANDS = (
    'alias', 'all_pixel', 'all_pixel_test', 'clear_cache', 'color', 'demo',
    'devices', 'info', 'load', 'list', 'remove', 'reset', 'run', 'save', 'set',
    'show', 'update')
MODULES = {c: import_module('bibliopixel.main.' + c) for c in COMMANDS}


def no_command(*_):
    log.printer('ERROR: No command entered')
    log.printer('Valid:', ', '.join(COMMANDS))
    return -1


def get_args(argv=sys.argv):
    argv = ['--help' if i == 'help' else i for i in argv[1:]]

    # argparse doesn't give command-specific help for `bp --help <command>`
    # so we use `bp <command> --help` (#429)
    if len(argv) == 2 and argv[0] == '--help':
        argv.reverse()

    try:
        argv.remove('--version')
    except:
        pass
    else:
        log.printer('BiblioPixel version %s' % common_flags.VERSION)
        if not argv:
            return

    if argv and not argv[0].isidentifier():
        # The first argument can't be a command so try to run it.
        argv.insert(0, 'run')

    if argv and argv[0].startswith('-'):
        log.printer(
            'bibliopixel: error: command line flags must appear at the end.')

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    for name, module in sorted(MODULES.items()):
        subparser = subparsers.add_parser(name, help=module.__doc__)
        common_flags.add_common_flags(subparser)
        module.set_parser(subparser)

    return parser.parse_args(argv)


def main():
    args = get_args()
    run = getattr(args, 'run', None)
    if not run:
        log.printer('ERROR: No command entered')
        log.printer('Valid:', ', '.join(COMMANDS))
        sys.exit(-1)

    aliases.ISOLATE = args.isolate

    if args.verbose and args.loglevel != 'frame':
        log.set_log_level('debug')
    else:
        log.set_log_level(args.loglevel)

    try:
        return run(args) or 0
    except Exception as e:
        if args.verbose:
            raise
        log.printer('ERROR:', e.args[0], file=sys.stderr)
        log.printer(*e.args[1:], sep='\n', file=sys.stderr)
        result = getattr(e, 'errorcode', -1)

    sys.exit(result)



import copy, multiprocessing
from . import server, trigger_process
from .. animation import STATE
from .. import collection
from ... project import load
from ... util import log
from bibliopixel.animation.off import OffAnim

DEFAULT_OFF = 'OFF_ANIM'
DEFAULT_ANIM_CONFIG = {
    'bgcolor': '#00ff00',
    'font_color': '#ffffff',
    'valid': True
}


BAD_DEFAULT_ERROR = """\
`{}` is not a valid default!
It must be one of the configured animation names."""


def normalize_name(name):
    return ''.join('_' if e is ' ' else e for e in name if e.isalnum() or e is ' ')


class RemoteControl(collection.Collection):
    @staticmethod
    def pre_recursion(desc):
        collection.Collection.pre_recursion(desc)

        animations = desc['animations']
        auto_demo = desc.pop('auto_demo', None)
        desc['anim_cfgs'] = []

        if auto_demo:
            auto_demo_run = auto_demo.pop('run', 10)  # default to 10 seconds per
            auto_demo['typename'] = 'sequence'
            auto_demo['data'] = dict(DEFAULT_ANIM_CONFIG, **auto_demo.get('data', {}))
            if 'name' not in auto_demo:
                auto_demo['name'] = 'DEMO_ANIM'
            auto_demo['data']['display'] = auto_demo['data'].get('display', auto_demo['name'])
            auto_demo['name'] = normalize_name(auto_demo['name'])
            auto_demo['animations'] = []
            animations.insert(0, {'animation': auto_demo, 'run': {'threaded': True}})

        # normalize name and display first
        for i, anim in enumerate(animations):
            anim['data'] = dict(DEFAULT_ANIM_CONFIG, **anim.get('data', {}))
            if not anim.get('name'):
                log.error('All animations should have the `name` parameter: {}'.format(anim))
                anim['name'] = str(i)
            anim['data']['display'] = anim['data'].get('display', anim['name'])
            anim['name'] = normalize_name(anim['name'])

            # This was originally done later as a side-effect.
            # Moved it here so it's obvious.  Seems like a defect - name is in
            # two places.
            anim['data']['name'] = anim['name']
            anim['run'] = anim.get('run', {})

            if auto_demo:
                demo_sub = copy.deepcopy(anim)
                demo_sub['run'].update(auto_demo_run)
                demo_sub['animation'].pop('name')
                demo_sub['animation'].pop('data')
                demo_sub['run']['threaded'] = False  # no threading allowed for internal sequences
                auto_demo['animations'].append(demo_sub)

            anim['run']['threaded'] = True  # threaded required
            desc['anim_cfgs'].append(anim['data'])

        return desc

    def __init__(self, *args, anim_cfgs,
                 external_access=False, port=5000,
                 title='BiblioPixel Remote', bgcolor='black',
                 font_color='white', default=None,
                 triggers=[], **kwds):
        super().__init__(*args, **kwds)
        self.internal_delay = 0  # never wait
        self.name_map = {}
        self.anim_cfgs = anim_cfgs

        for i, (anim, cfg) in enumerate(zip(self.animations, self.anim_cfgs)):
            name = cfg['name']
            if name in self.name_map:
                raise ValueError('Cannot have multiple animations with the same name: ' + name)
            if anim:
                anim.on_completion = self.on_completion
                self.name_map[name] = i
            else:
                # The animation failed to load.
                cfg.update(
                    valid=False,
                    display='FAILED: ' + cfg.get('display', '(no error)'),
                    bgcolor='rgb(48, 48, 48)',
                    font_color='white')
                self.name_map[name] = None

        if default is None:
            self.default = DEFAULT_OFF
        else:
            self.default = normalize_name(default)
            self.name_map[DEFAULT_OFF] = self.name_map[self.default]
            if self.default is None:
                self.default = DEFAULT_OFF

        if self.default == DEFAULT_OFF:
            off = OffAnim(self.layout)
            off.set_runner({'threaded': True})
            self.animations.append(off)
            self.name_map[DEFAULT_OFF] = len(self.animations) - 1

        self.index = self.name_map[DEFAULT_OFF]  # start with default animation

        self.ui_config = {
            'bgcolor': bgcolor,
            'font_color': font_color,
            'title': title
        }

        self.q_recv = multiprocessing.Queue()

        self.send_queues = {
            'RemoteServer': multiprocessing.Queue()
        }

        server_args = (
            external_access,
            port,
            self.send_queues['RemoteServer'],
            self.q_recv
        )

        self.server = multiprocessing.Process(target=server.run_server,
                                              args=server_args)

        self.handlers = {
            'run_animation': self.run_animation,
            'stop_animation': self.stop_animation,
            'get_config': self.get_config,
            'trigger_animation': self.run_animation,
            'brightness': self.change_brightness
        }

        self.trigger_procs = {}
        self.triggers = {}
        for trigger in triggers:
            typename = trigger.pop('typename')
            events = trigger.pop('events')
            if typename and events:
                load.code(typename)  # attempt early to fail early
                self.triggers.setdefault(typename, []).extend((events, trigger))
            else:
                raise ValueError('Triggers require `typename` and `events` fields!')

        for typename, trigger in self.triggers.items():
            events, kwargs = trigger
            self.trigger_procs[typename] = multiprocessing.Process(
                target=trigger_process.run_trigger,
                args=(typename, self.q_recv, events, kwargs))

    def cleanup(self, clean_layout=True):
        self.q_recv.close()
        for q in self.send_queues.values():
            q.close()
        self.server.terminate()
        for _, t in self.trigger_procs.items():
            t.terminate()
        super().cleanup(clean_layout)

    def on_completion(self, reason):
        if reason != STATE.canceled:
            self.stop_animation()

    # API Handlers
    def run_animation(self, name=None):
        if name is None:
            name = self.default

        if name not in self.name_map:
            error = 'Invalid animation name: {}'.format(name)
            log.info(error)
            return False, error

        if self.current_animation:
            self.current_animation.cleanup(clean_layout=False)
            self.index = -1

        log.info('Running animation: {}'.format(name))
        self.index = self.name_map[name]
        self.current_animation.start()
        return True, None

    def stop_animation(self, data):
        return self.run_animation()

    def change_brightness(self, data):
        if isinstance(data, str):
            try:
                data = int(data)
            except:
                return False, 'Invalid brightness value!'

        self.layout.set_brightness(data)
        return True, None

    def get_config(self, data):
        resp = {
            'ui': self.ui_config,
            'brightness': self.layout.brightness,
            'animations': list(self.anim_cfgs)
        }
        return True, resp

    def pre_run(self):
        self.server.start()
        for proc in self.trigger_procs.values():
            proc.start()

    def step(self, amt=None):
        self.run_animation()
        while not self.threading.stop_event.isSet():
            recv = self.q_recv.get()
            req = recv['req']
            if req not in self.handlers:
                resp = False, '{} is not a valid request!'.format(req)
            else:
                resp = self.handlers[req](recv['data'])

            resp_q = self.send_queues.get(recv['sender'])
            if resp_q:
                resp_q.put(resp)



import json, sys
from .. project import merge, defaults


def run(args):
    defaults.set_defaults(args.sections)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = DESCRIPTION
    parser.add_argument('sections', nargs='*')


DESCRIPTION = """
set -- set some or all sections of the project defaults.

Each argument to `bp default set` has to look like

  section=value

Example:

   bp set numbers=float dimensions=80
"""



import fractions
from . import control_source

try:
    import mido
    MESSAGE_TYPES = set(s['type'] for s in mido.messages.specs.SPECS)

except:
    mido = None
    MESSAGE_TYPES = set()


class Midi(control_source.ExtractedSource):
    """
    Midi is an ExtractorControlSource for MIDI messages.

    The input description is either a list of MIDI port names, or None, in
    which case all MIDI data from all MIDI ports is used.
    """

    # KEYS_BY_TYPE is a dictionary by message type of the fields we extract from
    # mido MIDI messages.
    # There are more mido message types that we haven't used yet.
    KEYS_BY_TYPE = {
        'aftertouch': ('channel', 'type', 'value'),
        'control_change': ('channel', 'type', 'control', 'value'),
        'note_off': ('channel', 'type', 'note', 'velocity'),
        'note_on': ('channel', 'type', 'note', 'velocity'),
        'pitchwheel': ('channel', 'type', 'pitch'),
        'program_change': ('channel', 'type', 'program'),
    }

    # Some numeric fields (channel, control, program and note) don't get
    # normalized because they are basically names.
    #
    NORMALIZERS = {
        'pitch': lambda x: fractions.Fraction(x - 8192) / 8192,
        'value': lambda x: fractions.Fraction(x) / 127,
        'velocity': lambda x: fractions.Fraction(x) / 127,
    }

    def __iter__(self):
        inputs = mido.get_input_names()
        for msg in mido.ports.MultiPort(mido.open_input(i) for i in inputs):
            yield vars(msg)



import errno
from . import log


ADDRESS_IN_USE_ERROR = """

Port {0} on your machine is already in use.
Perhaps BiblioPixel is already running on your machine?
"""

CACHED_SERVER_ERROR = """
Tried to open server of type {new_type} on port {port}, but there was already
a server of type {old_type} running there.
"""

CACHED_KWDS_WARNING = """
Cached server for port {server.port} had keywords {server.kwds},
but keywords {kwds} were requested.
"""


class CachedServer:
    def __init__(self, port, constructor, **kwds):
        self.server = self._make_server(port, constructor, **kwds)
        self.port = port
        self.constructor = constructor
        self.kwds = kwds

    def check_keywords(self, constructor, **kwds):
        if self.constructor != constructor:
            raise ValueError(CACHED_SERVER_ERROR.format(
                port=self.port,
                new_type=str(constructor),
                old_type=str(self.constructor)))

        if self.kwds != kwds:
            log.warning(CACHED_KWDS_WARNING.format(server=self, kwds=kwds))

    def close(self):
        pass

    def __getattr__(self, key):
        # Pass through all other attributes to the server.
        return getattr(self.server, key)

    @staticmethod
    def _make_server(port, constructor, **kwds):
        try:
            return constructor(port, **kwds)
        except OSError as e:
            if e.errno == errno.EADDRINUSE:
                e.strerror += ADDRESS_IN_USE_ERROR.format(port)
                e.args = (e.errno, e.strerror)
            raise


class ServerCache:
    ENABLE = True

    def __init__(self, **kwds):
        self.servers = {}
        self.kwds = kwds

    def get_server(self, port, **kwds):
        kwds = dict(self.kwds, **kwds)

        if not self.ENABLE:
            return CachedServer._make_server(port, **kwds)

        server = self.servers.get(port)
        if server:
            server.check_keywords(**kwds)
        else:
            server = CachedServer(port, **kwds)
            self.servers[port] = server

        return server

    def close(self, port):
        server = self.servers.pop(port)
        server.server.close()



"""
Run a project description file.
"""

import os, sys, time, traceback
from . import common_flags, simpixel
from .. util import log
from .. animation.collection import Collection
from .. project import load

RUN_ERROR = """When reading description:
{desc}
encountered exception
{exception}
"""

FAILURE_ERROR = '{count} project{s} failed'


def _load_py(filename):
    module = load.module(filename)
    project = module.get('PROJECT', {})

    if 'animation' not in project:
        no_extension = os.path.splitext(filename)[0]
        module_name = os.path.split(no_extension)[-1]
        names = module.keys()
        try:
            animation_name = load.guess_name(names, module_name, filename)
        except:
            if 'Animation' not in names:
                raise ValueError('Cannot deduce animation in file ' + filename)
            animation_name = 'Animation'
        animation = module[animation_name]
        if not callable(animation):
            raise ValueError('Animation "%s" in file "%s" is not callable'
                             % (animation_name, filename))

        project['animation'] = {'datatype': animation}

    return project


def _get_animations(args):
    animations, failed = [], []

    for filename in args.name:
        saved_path = sys.path[:]
        desc = '(not loaded)'
        try:
            if filename.endswith('.py'):
                desc = _load_py(filename)
            else:
                desc = load.data(filename)
            animation = common_flags.make_animation(args, desc)
            animations.append(animation)

        except Exception as exception:
            if filename.endswith('.py'):
                raise
            if args.verbose:
                exception = traceback.format_exc()
            msg = RUN_ERROR.format(**locals())
            failed.append((msg, exception.args))

        finally:
            sys.path[:] = saved_path

    if not failed:
        return animations

    log.error(FAILURE_ERROR.format(
        count=len(failed), s='' if len(failed) == 1 else 's'))
    for msg, args in failed:
        log.error(msg + '\n' + '\n'.join(str(a) for a in args))
    raise ValueError('Run aborted')


def _run_animations(animations, pause, names):
    needs_pause = False
    assert len(animations) == len(names)
    for animation, name in zip(animations, names):
        log.debug('Running file %s', name)
        if needs_pause:
            pause and time.sleep(float(pause))
        else:
            needs_pause = True

        try:
            animation.layout.start()
            animation.start()
        except KeyboardInterrupt:
            log.warning('\nTermination requested by user.')
            needs_pause = False

        animation.cleanup()
        animation.layout.cleanup_drivers()


def run(args):
    if args.fail_on_exception:
        Collection.FAIL_ON_EXCEPTION = True

    args.name = args.name or ['']
    animations = _get_animations(args)

    if args.simpixel:
        simpixel.open_simpixel(args.simpixel)
    elif args.s:
        simpixel.open_simpixel()

    _run_animations(animations, args.pause, args.name)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = 'Run specified BiblioPixel project from file or URL.'

    common_flags.add_project_flags(parser)

    parser.add_argument(
        'name', nargs='*',
        help='Path project files - can be a URL or file system location')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Enter JSON directly as a command line argument.')



from .. project import defaults


def run(args):
    defaults.reset_defaults(args.sections)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = DESCRIPTION
    parser.add_argument('sections', nargs='*')


DESCRIPTION = """reset -- reset some or all sections of the project defaults.
Example:

    # Reset the layout and animation sections
    $ bp default reset layout animation

    # Reset all sectionsn
    $ bp default reset
"""



import json, os
from .. project import data_maker, defaults, project

"""Common command line arguments for run and demo."""


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(dirname(dirname(abspath(
        __file__)))), 'bibliopixel', 'VERSION')
    return open(filename).read().strip()


VERSION = _get_version()
COMPONENTS = 'driver', 'layout', 'animation'
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
LOG_LEVELS = ('debug', 'info', 'warning', 'error', 'critical')
ENABLE_PRESETS = False
NUMBER_TYPES = ('python',) + data_maker.NUMPY_TYPES


def add_common_flags(parser):
    parser.add_argument(
        '--loglevel', choices=LOG_LEVELS, default='info', help=LOGLEVEL_HELP)
    parser.add_argument(
        '--verbose', '-v', action='store_true', help=VERBOSE_HELP)
    parser.add_argument(
        '--version', action='store_true', help=VERSION_HELP)
    parser.add_argument(
        '-x', '--isolate', action='store_true', help=ISOLATE_HELP)


def add_project_flags(parser):
    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')

    parser.add_argument(
        '-b', '--brightness', default=None,
        help='Override project brightness value')

    parser.add_argument(
        '-d', '--defaults', default=None, nargs='*',
        action='append', help='Use this default setting')

    parser.add_argument(
        '--dimensions', '--dim', default=None,
        help='x, (x, y) or (x, y, z) dimensions for project')

    parser.add_argument(
        '-f', '--fail_on_exception', action='store_true',
        help='If true, bp fail if any subanimation fails to construct')

    parser.add_argument(
        '-l', '--layout', default=None,
        help='Default layout class if no layout is specified')

    parser.add_argument(
        '--numbers', '-n', default='python', choices=NUMBER_TYPES,
        help=NUMBERS_HELP)

    parser.add_argument('-p', '--path', default=None, help=PATH_HELP)

    parser.add_argument(
        '--pause', default=0, help='Time to pause between running animations')

    parser.add_argument(
        '-r', '--run_for', default=None, help=RUN_FOR_HELP)

    parser.add_argument(
        '-s', action='store_true', help='Run SimPixel at the default URL')

    parser.add_argument(
        '--simpixel', help='Run SimPixel at a specific URL')

    parser.add_argument(
        '--slideshow', default=0, help='Run `bp` sequence in slideshow mode, '
        'where each animation is shown for this many seconds')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')


def _make_project_flags(args):
    def get_value(name):
        value = getattr(args, name, None)
        if not value:
            return {}

        if '{' in value:
            return json.loads(value)

        return {'typename': value}

    project_flags = {name: get_value(name) for name in COMPONENTS}
    if args.ledtype:
        project_flags['driver']['ledtype'] = args.ledtype

    if args.brightness:
        project_flags['layout']['brightness'] = int(args.brightness)

    if args.defaults:
        # Arguments come in like: [['foo'], ['bar'], ['baz'], ...]
        # I have no idea why.
        defs = [(a[0] if isinstance(a, list) else a) for a in args.defaults]
        defaults.set_project_defaults(defs)

    if args.numbers != 'python':
        project_flags['numbers'] = args.numbers

    if args.run_for is not None:
        project_flags.setdefault('run', {})['seconds'] = float(args.run_for)

    if args.slideshow:
        project_flags.setdefault('animation', {})['slideshow'] = float(
            args.slideshow)

    if args.dimensions is not None:
        dimensions = args.dimensions.split(',')
        try:
            project_flags['dimensions'] = [int(i) for i in dimensions]
        except:
            raise ValueError('--dimensions must be one to three numbers '
                             'separated by commas.')

    return project_flags


def make_project(args, desc):
    project_flags = _make_project_flags(args)
    return project.project(project_flags, desc)


def make_animation(args, desc):
    return make_project(args, desc).make_animation()


# Help messages.

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use loady-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

LOGLEVEL_HELP = """\
Set what level of events to log. Higher log levels print less."""

ISOLATE_HELP = """\
Run BiblioPixel in isolated mode, where it cannot see your local files.
This means that it will not see any local Python classes in your directories
and it won't see your local aliases.

Running your project in isolated mode help makes sure that your project will
work on other machines without modification.
"""

NUMBERS_HELP = """
The type of numbers that are used in color list calculations.

  `python` means to use the classic Python lists of (r, g, b) tuples.

  Anything else is a numpy type, which means that bp uses numpy arrays, which
  use much faster arithmetic.

  numpy types include:
    """ + ' '.join(data_maker.NUMPY_TYPES)

PRESET_HELP = """Filenames for preset library"""

RUN_FOR_HELP = """How long to run the animation (overrides runner.seconds)."""

VERBOSE_HELP = """\
If this is set, then errors are reported with a full stack trace, and
loglevel is by default set to debug..
"""

VERSION_HELP = """\
Print the current version number of BiblioPixel (%s).
""" % VERSION


# TODO: this should go somewhere else
"""
Set defaults for projects using JSON.

A Project is a JSON file or equivalently a Python dictionary that describes
a BibliopPixel installation.  The top-level keys in a Project are called the
sections and a Project might have the following sections:

    {sections}

Often some sections of your project correspond to hardware and thus rarely
change, so the `bp default` command allows you to set defaults so you
don't have to mention these from your project at all.
"""



"""
To get a color from a name, use

    COLORS.colorname

or if the name is a variable or has a space in it,

    COLORS['color name']

To get a name from a color, use

    COLOR((0, 255, 255))
"""

import re
from . import juce, classic


def _to_triplet(color):
    rg, b = color // 256, color % 256
    r, g = rg // 256, rg % 256
    return r, g, b


def _classic_pairs():
    find = re.compile(r'[A-Z][a-z0-9]+').finditer

    for name in dir(classic):
        if name == 'Green_HTML':
            yield 'green html', classic.Green_HTML

        elif len(name) > 1 and name[0].isupper() and name[1].islower():
            key = ' '.join(i.group(0).lower() for i in find(name))
            yield key, getattr(classic, name)


def _from_number(s):
    s = s.strip()
    for prefix in '0x', '#':
        if s.startswith(prefix):
            return int(s[len(prefix):], 16)

    return int(s)


JUCE_COLORS = {k: _to_triplet(v) for k, v in juce.COLORS.items()}
CLASSIC_COLORS = dict(_classic_pairs())
COLOR_DICT = dict(CLASSIC_COLORS, **JUCE_COLORS)

SECONDARY_NAMES = juce.SECONDARY_NAMES.union({'off', 'on'})
TO_NAME = {v: k for k, v in COLOR_DICT.items() if k not in SECONDARY_NAMES}
TO_COLOR = {k.replace(' ', ''): v for k, v in COLOR_DICT.items()}


def name_to_color(name):
    def to_color(name):
        name = name.lower()
        if ',' in name:
            if name.startswith('(') and name.endswith(')'):
                name = name[1:-1]

            r, g, b = name.split(',')
            return _from_number(r), _from_number(g), _from_number(b)

        try:
            n = _from_number(name)
        except:
            return TO_COLOR[name.replace(' ', '').lower()]

        return _to_triplet(n)

    try:
        color = to_color(name)
    except:
        raise ValueError('Unknown color name %s' % name)

    if not all(0 <= i <= 255 for i in color):
        raise ValueError('Component out of range: %s' % color)

    return color


def color_to_name(color):
    try:
        return TO_NAME[tuple(color)]
    except:
        return str(tuple(color))


def toggle(s):
    """Toggle between a name and a tuple representation."""
    is_numeric = ',' in s or s.startswith('0x') or s.startswith('#')
    c = name_to_color(s)
    return color_to_name(c) if is_numeric else str(c)


class Colors(object):
    def __getitem__(self, name):
        try:
            return name_to_color(name)
        except:
            raise KeyError(name)

    def __getattr__(self, name):
        try:
            return name_to_color(name)
        except:
            raise AttributeError("COLORS has no attribute' %s'" % name)

    def __call__(self, color):
        return color_to_name(color)


COLORS = Colors()



"""
An address identifies the location data within a Python object.

An address looks like

    foo.bar[32][5][baz].bang

That address would mean "given some object x, the value of
x.foo.bar[32][5]['baz'].bang".

Addresses are divided into "segments".  Segments in brackets are indexes or keys;
otherwise they are attributes.  In the example above, the segments are
foo, bar, [32], [5], [baz] and bang; foo and bar are attributes.
baz is a string index, and 32 and 5 are numeric indexes.

You can use addresses to either get or set values in an object. Getting or
setting throw a ValueError if the address can't be get or set, so calling code
that needs to be robust even in the face of bad user data will have to catch and
report or handle the error.

Numerical indexes are automatically distinguished from string indexes.  This is
convenient for the user but prevents them from having dictionaries like
{1: 'number one', '1': 'string one'} - they probably didn't want to do that
anyway.
"""


class Address:
    class Segment:
        def __init__(self, name):
            self.name = name

    class Attribute(Segment):
        def get(self, x):
            return getattr(x, self.name)

        def set(self, x, value):
            return setattr(x, self.name, value)

    class Index(Segment):
        def get(self, x):
            return x[self.name]

        def set(self, x, value):
            x[self.name] = value

    def __init__(self, s):
        def generate():
            # First split on dots, then use [ and ] to split out indices.
            for i, part in enumerate(s.strip().split('.')):
                head, *rest = part.split('[')

                # Suppose our address was 'xxx[yyy][zzz]'
                # Now we have first='xxx' and rest = 'yyy]', 'zzz]'
                if head:
                    yield Address.Attribute(head)
                elif i:
                    # An index [] is only allowed to start the *first* segment -
                    # for example, an address like a.[2] is forbidden.
                    raise ValueError

                for r in rest:
                    # A segment must have exactly one ']', exactly at the end.
                    index, between = r.split(']')
                    if between:
                        raise ValueError

                    try:
                        index = int(index)
                    except ValueError:
                        pass

                    yield Address.Index(index)

        try:
            self.address = list(generate())
        except:
            raise ValueError('%s is not a legal address' % s)

        if not self.address:
            raise ValueError('Empty Addresses are not allowed')

    @staticmethod
    def _get(x, address):
        for a in address:
            x = a.get(x)
        return x

    def get(self, x):
        return self._get(x, self.address)

    def set(self, x, value):
        *first, last = self.address
        parent = self._get(x, first)
        last.set(parent, value)



"""
Converts its arguments between color names and color tuples.
"""

from .. project import aliases, alias_lists
from .. util import log

DELETE_ALL_PROMPT = 'Delete all aliases - are you sure? (yN) '


def _assign(parts):
    def assign(i):
        name, equals, value = parts[i:i + 3]
        if equals != '=' or not name.isidentifier():
            raise ValueError()

        return name, value

    # Compute all the assignments first to fail early.
    assignments = [assign(i) for i in range(0, len(parts), 3)]
    for name, value in assignments:
        alias_lists.set_alias(name, value)
        log.printer('Set:', name, '=', value)


def run(args):
    if args.delete_all:
        if args.aliases:
            raise ValueError('--delete-all takes no arguments')

        if not input(DELETE_ALL_PROMPT).lower().startswith('y'):
            raise ValueError('--delete-all aborted')

        alias_lists.delete_all_alias()
        log.printer('All aliases deleted.')
        return

    if not args.aliases:
        alias_lists.print_aliases(args.builtin, args.by_value)
        return

    parts = ' = '.join(' '.join(args.aliases).split('=')).split()

    if '=' in parts:
        if args.delete:
            raise ValueError('Cannot --delete %s' % ' '.join(args.aliases))
        try:
            _assign(parts)
        except Exception as e:
            raise ValueError('Bad assignment %s' % ' '.join(args.aliases))

    elif args.delete:
        success = []
        for alias in parts:
            try:
                alias_lists.delete_alias(alias)
                success.append(alias)
            except:
                log.printer('Alias', alias, 'did not exist')
        success and log.printer('Deleted aliases:', *success)

    else:
        for alias in parts:
            value = alias_lists.get_alias(alias, aliases.ISOLATE)
            alias_lists.print_alias(alias, value)


def set_parser(parser):
    parser.set_defaults(run=run)

    parser.add_argument(
        'aliases', nargs='*',
        help='Names of aliases',
        default='')

    parser.add_argument(
        '--builtin', action='store_true',
        help='List the builtin aliases')

    parser.add_argument(
        '--by-value', action='store_true',
        help='Sort aliases by value, not name')

    parser.add_argument(
        '--delete', action='store_true',
        help='Delete one or more built-in aliases')

    parser.add_argument(
        '--delete-all', action='store_true',
        help='Delete all built-in aliases')



import collections
from . import construct

# Elements in the DEFAULT_PROJECT are listed in order of dependency: sections
# can only be dependent on sections that are above them in this list.
# For example, animation is at the bottom which are  dependent on run and
# layout, which depend on drivers, and everything depends on path and typename.
DEFAULT_PROJECT = collections.OrderedDict((
    ('aliases', {}),
    ('path', ''),
    ('typename', 'bibliopixel.project.project.Project'),
    ('numbers', 'python'),
    ('maker', 'bibliopixel.project.data_maker.Maker'),
    ('driver', {}),
    ('drivers', []),
    ('dimensions', ()),
    ('layout', {}),
    ('run', {}),
    ('animation', {}),
))
PROJECT_SECTIONS = tuple(DEFAULT_PROJECT.keys())

NOT_MERGEABLE = (
    'datatype', 'dimensions', 'drivers', 'numbers', 'path', 'typename')

SECTION_ISNT_DICT_ERROR = 'Project section "%s" is %s, should be dictionary'


def merge(*projects):
    """
    Merge zero or more dictionaries representing projects with the default
    project dictionary and return the result
    """
    result = {}
    for project in projects:
        for name, section in (project or {}).items():
            if name in NOT_MERGEABLE:
                result[name] = section
            elif section and not isinstance(section, (dict, str)):
                cname = section.__class__.__name__
                raise ValueError(SECTION_ISNT_DICT_ERROR % (name, cname))
            else:
                result.setdefault(name, {}).update(construct.to_type(section))

    return result



from . animation import BaseAnimation
from . circle import BaseCircleAnim
from . game import BaseGameAnim
from . matrix import BaseMatrixAnim
from . cube import BaseCubeAnim
from . off import OffAnim
from . sequence import Sequence
from . strip import BaseStripAnim
from . receiver import BaseReceiver
from . tests import StripChannelTest, MatrixChannelTest, MatrixCalibrationTest



from .. project import defaults


def run(args):
    defaults.list_saved_defaults()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = 'List all saved defaults settings'



import functools, threading


class Loop(threading.Thread):
    def __init__(self, loop=None, daemon=True, **kwds):
        super().__init__(daemon=daemon, **kwds)
        self.running = True
        self.loop = loop or self.loop

    def run(self):
        while self.running:
            self.loop()

    def stop(self):
        self.running = False


class Loops(object):
    def __init__(self, loops):
        self.loops = list(loops)

    def start(self):
        map(lambda x: x.start(), self.loops)

    def stop(self):
        map(lambda x: x.stop(), self.loops)



from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util.colors import gamma as _gamma
from . import interfaces


class SPIBase(DriverBase):
    """Base driver for controling SPI devices on systems like the Raspberry Pi
       and BeagleBone"""
    def __init__(self, num, dev='/dev/spidev0.0',
                 spi_interface=None, spi_speed=1,
                 interface='FILE',  # DEPRECATED
                 **kwargs):
        super().__init__(num, **kwargs)
        from ...project.types.spi_interface import make
        # See https://github.com/ManiacalLabs/BiblioPixel/issues/419

        maker = interfaces._SPI_INTERFACES[make(spi_interface or interface)]
        self._interface = maker(dev=dev, spi_speed=spi_speed)

    def _send_packet(self):
        self._interface.send_packet(self._packet)

    def _compute_packet(self):
        self._render()
        self._packet = self._interface.compute_packet(self._buf)



import copy, json


class DataFile(object):
    def __init__(self, filename, open=open):
        self.filename = filename
        self.open = open
        self.read()

    def read(self):
        try:
            fp = self.open(self.filename)
        except FileNotFoundError:
            self.data = {}
        else:
            self.data = json.load(fp)

    def write(self):
        json.dump(self.data, self.open(self.filename, 'w'))

    def get(self, key):
        return self.data.get(key)

    def set(self, key, value):
        self.data[key] = value
        self.write()

    def set_items(self, items):
        for k, v in items:
            self.data[k] = v
        self.write()

    def delete(self, key):
        del self.data[key]
        self.write()

    def delete_all(self):
        self.data.clear()
        self.write()



import operator, os
from .. util import datafile, log

USER_ALIAS_FILE = os.path.expanduser('~/.bibliopixel_aliases')
USER_ALIASES = datafile.DataFile(USER_ALIAS_FILE)
PROJECT_ALIASES = {}


BUILTIN_ALIASES = {
    # drivers
    'apa102': 'bibliopixel.drivers.SPI.APA102.APA102',
    'sk9822': 'bibliopixel.drivers.SPI.APA102.APA102',
    'dummy': 'bibliopixel.drivers.dummy_driver.Dummy',
    'hue': 'bibliopixel.drivers.hue.Hue',
    'lpd8806': 'bibliopixel.drivers.SPI.LPD8806.LPD8806',
    'mirror': 'bibliopixel.drivers.mirror.MirrorDriver',
    'network': 'bibliopixel.drivers.network.Network',
    'network_udp': 'bibliopixel.drivers.network_udp.NetworkUDP',
    'serial': 'bibliopixel.drivers.serial.Serial',
    'simpixel': 'bibliopixel.drivers.SimPixel.SimPixel',
    'ws281x': 'bibliopixel.drivers.SPI.WS281X.WS281X',
    'ws2801': 'bibliopixel.drivers.SPI.WS2801.WS2801',
    'spi': 'bibliopixel.drivers.SPI.SPI',
    'pi_ws281x': 'bibliopixel.drivers.PiWS281X.PiWS281X',

    # layouts
    'circle': 'bibliopixel.layout.circle.Circle',
    'cube': 'bibliopixel.layout.cube.Cube',
    'matrix': 'bibliopixel.layout.matrix.Matrix',
    'pov': 'bibliopixel.layout.pov.POV',
    'strip': 'bibliopixel.layout.strip.Strip',

    # animations
    'off': 'bibliopixel.animation.off.OffAnim',
    'feedback': 'bibliopixel.animation.feedback.Feedback',
    'matrix_calibration': 'bibliopixel.animation.tests.MatrixCalibrationTest',
    'matrix_test': 'bibliopixel.animation.tests.MatrixChannelTest',
    'mixer': 'bibliopixel.animation.mixer.Mixer',
    'receiver': 'bibliopixel.animation.receiver.BaseReceiver',
    'reprocess': 'bibliopixel.animation.reprocess.reprocess.Reprocess',
    'remote': 'bibliopixel.animation.remote.control.RemoteControl',
    'sequence': 'bibliopixel.animation.Sequence',
    'strip_test': 'bibliopixel.animation.tests.StripChannelTest',
}


def get_alias(alias, isolate=False):
    return (
        PROJECT_ALIASES.get(alias) or
        (not isolate and USER_ALIASES.get(alias)) or
        BUILTIN_ALIASES.get(alias))


def print_alias(alias, value, printer=log.printer):
    if value:
        printer('%s = %s' % (alias, value))
    else:
        printer('# %s is not defined.' % alias)


def print_aliases(builtin, by_value=False, printer=log.printer):
    """
    Args:
        by_value: sort either by alias name, or by alias value
    """

    aliases = BUILTIN_ALIASES if builtin else USER_ALIASES.data
    if not aliases:
        printer('(no aliases)')
        return

    key_func = operator.itemgetter(int(by_value))
    for alias, value in sorted(aliases.items(), key=key_func):
        print_alias(alias, value, print)


set_alias = USER_ALIASES.set
delete_alias = USER_ALIASES.delete
delete_all_alias = USER_ALIASES.delete_all
